/* Meadow Install Program */
#include <windows.h>
#include <ddeml.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include "../src/mw32reg.h"

LPCTSTR meadow_version_string = MEADOW;
LPCTSTR meadow_common_root = REG_COMMON_ROOT;
LPCTSTR meadow_common_env_root = REG_COMMON_ENV_ROOT;
LPCTSTR meadow_version_root = REG_VERSION_ROOT;
LPCTSTR meadow_version_env_root = REG_VERSION_ENV_ROOT;

int w32_platform_id;
int codepage;

typedef struct slmessage
{
  int cp;
  unsigned char *message;
} slmessage;

/* US CP is 437,.... -1 must be last entry. It is default...*/

#define ENGLISH -1
#define NIHONGO 932

slmessage access_denied[] =
{
  {NIHONGO, "ANZXۂ܂\n"
   "AdministratorŎsĂB\n"},
  {ENGLISH, "Access denied.\n"
   "Run this program with administrator privilege."}
};
slmessage undetect_error[] =
{
  {NIHONGO, "VXěs̃G[\n"},
  {ENGLISH, "Undetected system error.\n"},
};
slmessage usage[] =
{
  {ENGLISH, "Usage:install [/c] [/a] [/A] [HOME-DIRECTORY]\n"},
};
slmessage lt_check_error[] =
{
  {NIHONGO, "WJɎsĂ܂B\n"
   "INSTALL.Meadow WinZip̒ӓ_ǂǂ݂ȂB\n"},
  {ENGLISH, "Some files not extracted correctly from the archive.\n"
   "Read the description about WinZip in INSTALL.Meadow.\n"},
};
slmessage lfn_check_error[] =
{
  {NIHONGO, "WJɎsĂ܂B\n"
   "Long File Name̎gtar & gzip gĉB\n"},
  {ENGLISH, "Some files not extracted correctly from the archive.\n"
   "Use tar & gzip which can handle Long File Names properly.\n"},
};
slmessage pf_error[] =
{
  {NIHONGO, "̃vOWindows95/NTł̂ݓ삵܂B\n"},
  {ENGLISH, "This program can run only on Windows95 or NT.\n"},
};
slmessage nd_homedrive[] =
{
  {NIHONGO, "HOMEDRIVE`ĂȂB\n"},
  {ENGLISH, "HOMEDRIVE is not defined. \n"},
};
slmessage nd_homepath[] =
{
  {NIHONGO, "HOMEPATH`ĂȂB\n"},
  {ENGLISH, "HOMEPATH is not defined. \n"},
};
slmessage home_prompt[] =
{
  {NIHONGO, "fBtHg̃z[fBNg͂ĂB\n"
   "ݒ肵ȂƂ̏ꏊMeadow.emacsǂ݂܂B\n"},
  {ENGLISH, "Enter the default home directory. (system default)\n"
   "Meadow read .emacs in this directory without any settings.\n"},
};
void slmessage_fprintf(lang, stream, slfmt, ...)
     int lang;
     FILE *stream;
     slmessage *slfmt;
{
  va_list va;

  va_start(va, slfmt);
  for (;!(slfmt->cp == lang || slfmt->cp == -1);slfmt++);
  vfprintf(stream, slfmt->message, va);
  va_end(va);
}

HDDEDATA CALLBACK DdeCallback (UINT uType, UINT uFmt, HCONV hconv,
			       HSZ hsz1, HSZ hsz2, HDDEDATA hdata,
			       DWORD dwData1, DWORD dwData2)
{
  return NULL;
}

#define DdeCommand(str) 	\
	DdeClientTransaction (str, strlen (str)+1, HConversation, (HSZ)NULL, \
		              CF_TEXT, XTYP_EXECUTE, 30000, NULL)

#define MULE_CLASS ""

void error_op(int errorflag)
{
  if (errorflag != ERROR_SUCCESS){
    switch(errorflag) {
    case ERROR_ACCESS_DENIED:
      slmessage_fprintf(codepage, stderr, access_denied);
      break;
    default:
      slmessage_fprintf(codepage, stderr, undetect_error);
      break;
    }
    exit(100);
  }
  return;
}

void error_usage(void)
{
  slmessage_fprintf(codepage, stderr, usage);
  exit(1);
}

void init_current_environment()
{
  int version;

  codepage = GetConsoleCP();

  version = GetVersion();
  if ( 0 == (version & 0x80000000) )
    w32_platform_id = 0;
  else if ((version & 0xC0000000) == 0xC0000000)
    w32_platform_id = 1;
  else {
    slmessage_fprintf(codepage, stderr, pf_error);
    exit(1);
  }
}

HKEY create_key(HKEY hrootkey, LPCTSTR entry)
{
  HKEY hkey;
  DWORD result;
  int errorflag;

  errorflag = RegCreateKeyEx(hrootkey, entry,
			     0, MULE_CLASS, REG_OPTION_NON_VOLATILE,
			     KEY_WRITE, NULL, &hkey, &result);

  error_op(errorflag);

  return hkey;
}

void check_files(unsigned char *mule_root)
{
  unsigned char buf[MAX_PATH];
  FILE *handle;
  int data, flag;

  sprintf(buf, "%slisp\\mw32scroll.el", mule_root);
  flag = access(buf, 0);
  if (flag != 0) {
    slmessage_fprintf(codepage, stderr, lfn_check_error);
    exit(100);
  }
  sprintf(buf, "%slisp\\loaddefs.el", mule_root);
  handle = fopen(buf, "rb");
  while ((data = fgetc(handle)) != EOF) {
    if (data == '\n') break;
    if (data == '\r') {
      slmessage_fprintf(codepage, stderr, lt_check_error);
      exit(100);
    }
  }
  return;
}

void set_value(HKEY hkey, unsigned char *name, unsigned char *value)
{
  int errorflag;

  errorflag = RegSetValueEx (hkey, name, 0, REG_SZ,
			     value, lstrlen (value) + 1);
  error_op(errorflag);
}

void set_value_path(HKEY hkey, 
		    unsigned char *name,
		    unsigned char *dir_name,
		    unsigned char *subdir)
{
  unsigned char full_name[MAX_PATH];
  strcpy(full_name, dir_name);
  strcat(full_name, subdir);
  set_value(hkey, name, full_name);
}

void get_root(unsigned char *program, unsigned char *result)
{
  unsigned char dir_name[_MAX_DIR];
  unsigned char drive_name[_MAX_DRIVE];
  unsigned char file_name[_MAX_FNAME];
  unsigned char ext_name[_MAX_EXT];

  _splitpath(program, drive_name, dir_name, file_name, ext_name);

  sprintf(result, "%s%s", drive_name, dir_name);
}

void get_shell_name(unsigned char *result)
{
  unsigned char dir_name[_MAX_DIR];
  unsigned char drive_name[_MAX_DRIVE];
  unsigned char file_name[_MAX_FNAME];
  unsigned char ext_name[_MAX_EXT];
  unsigned char *shell_name;

#if 0
  shell_name = getenv("COMSPEC");
  if (!shell_name) {
    if (w32_platform_id == 1)
      strcpy(result, "cmdproxy");
    else
      strcpy(result, "cmdproxy");
  } else {
    _splitpath(shell_name, drive_name, dir_name, file_name, ext_name);
    sprintf(result, "%s%s", file_name, ext_name);
  }
#else
  strcpy(result, "cmdproxy");
#endif
}

void set_registry_data(int personalp,
		       int versionp,
		       char *main_root,
		       char *shell_name,
		       char *home_dir)
{
  HKEY hkey;
  TCHAR tmpstr[MAX_PATH];

  if (versionp)
    sprintf(tmpstr, meadow_version_env_root, meadow_version_string);
  else
    strcpy(tmpstr, meadow_common_env_root);

  if (personalp)
    {
      hkey = create_key(HKEY_CURRENT_USER, tmpstr);
      set_value(hkey, "HOME", home_dir);
    }
  else
    {
      unsigned char buffer[2048];
      unsigned char main_top_root[MAX_PATH];
      int i;

      i = strlen(main_root) - 2;
      for (;i >= 0;i--) if (main_root[i] == '\\') break;

      if (i > 0)
	{
	  memcpy(main_top_root, main_root, i + 1);
	  main_top_root[i + 1] = '\0';
	}
      else
	strcpy(main_top_root, main_root);

      hkey = create_key(HKEY_LOCAL_MACHINE, tmpstr);
      sprintf(buffer, "%ssite-lisp;%ssite-lisp;%slisp;%sleim",
	      main_root, main_top_root, main_root, main_root);
      set_value(hkey, "EMACSLOADPATH", buffer);
      set_value_path(hkey, "EMACSDATA", main_root, "etc");
      set_value_path(hkey, "EMACSPATH", main_root, "bin");
      set_value_path(hkey, "EMACSLOCKDIR", main_root, "lock");
      set_value_path(hkey, "INFOPATH", main_root, "info");
      set_value_path(hkey, "EMACSDOC", main_root, "etc");
      set_value(hkey, "TERM", "CMD");
      set_value(hkey, "SHELL", shell_name);
      set_value(hkey, "HOME", home_dir);
    }
  RegCloseKey(hkey);
}

void remove_registry_data(int personalp, int versionp)
{
  HKEY hkey;

  if (personalp)
    {
      if (versionp)
	RegDeleteKey(HKEY_CURRENT_USER, meadow_version_root);
      else
	RegDeleteKey(HKEY_CURRENT_USER, meadow_common_root);
    }
  else
    {
      if (versionp)
	RegDeleteKey(HKEY_LOCAL_MACHINE, meadow_version_root);
      else
	RegDeleteKey(HKEY_LOCAL_MACHINE, meadow_common_root);
    }
}

void call_dumpprog(unsigned char *main_root)
{
  unsigned char dumpprog[MAX_PATH];

  sprintf(dumpprog, "%sbin", main_root);
  chdir(dumpprog);

  if (w32_platform_id == 1)
    sprintf(dumpprog, "%sbin\\dump95.bat", main_root);
  else
    sprintf(dumpprog, "%sbin\\dumpnt.bat", main_root);

  system(dumpprog);
}

void register_to_shell(char *main_root)
{
  DWORD idDde;
  HCONV HConversation;
  HSZ program_manager;
  char additem[MAX_PATH*2 + 100];

  idDde = 0;
  DdeInitialize (&idDde, (PFNCALLBACK)DdeCallback, APPCMD_CLIENTONLY, 0);

  program_manager = DdeCreateStringHandle (idDde, "PROGMAN", CP_WINANSI);

  if (HConversation = DdeConnect (idDde, program_manager,
				  program_manager, NULL))
    {
      DdeCommand ("[CreateGroup (Meadow)]");
      DdeCommand ("[ReplaceItem (Meadow)]");
      sprintf (additem, "[AddItem (%s\\bin\\meadow.exe,Meadow,"
	       "%s\\bin\\meadow.ico)]",
	       main_root, main_root);
      DdeCommand (additem);

      DdeDisconnect (HConversation);
    }

  DdeFreeStringHandle (idDde, program_manager);

  DdeUninitialize (idDde);
}

int check_opt(char *arg, char *opt)
{
  if ((arg[0] != '/')
      && (arg[0] != '-')) return 0;

  if (strcmp(&arg[1], opt) == 0) return 1;
  return 0;
}

int main (int argc, char **argv)
{
  unsigned char main_root[MAX_PATH];
  unsigned char this_prog[MAX_PATH];
  unsigned char home_dir[MAX_PATH];
  unsigned char shell_name[MAX_PATH];
  unsigned char *p;
  int personal_flag = 0;
  int version_specific_flag = 1;
  int remove_flag = 0;

  init_current_environment();

  GetModuleFileName(NULL, this_prog, MAX_PATH);
  get_root(this_prog, main_root);

  check_files(main_root);

  {
    int i;
    int set_home_dir = 0;

    for (i = 1;i < argc;i++)
      {
	if (check_opt(argv[i], "a"))
	  {
	    p = getenv("HOMEDRIVE");
	    if (!p)
	      {
		slmessage_fprintf(codepage, stderr, nd_homedrive);
		exit(1);
	      }
	    strcpy(home_dir, p);
	    p = getenv("HOMEPATH");
	    if (!p)
	      {
		slmessage_fprintf(codepage, stderr, nd_homepath);
		exit(1);
	      }
	    strcat(home_dir, p);
	    set_home_dir = 1;
	    personal_flag = 1;
	  }
	if (check_opt(argv[i], "A"))
	  {
	    personal_flag = 1;
	    set_home_dir = -1;
	  }
	if (check_opt(argv[i], "c"))
	  {
	    version_specific_flag = 0;
	  }
	if (check_opt(argv[i], "u"))
	  {
	    remove_flag = 1;
	    set_home_dir = 1;
	  }
	else
	  {
	    if (i != (argc - 1))
	      error_usage();
	    strncpy(home_dir, argv[i], MAX_PATH);
	    set_home_dir = 1;
	  }
      }
    if (set_home_dir < 0)
      error_usage();
    else if (set_home_dir == 0)
      {
	slmessage_fprintf(codepage, stdout, home_prompt);
	fgets(home_dir, MAX_PATH, stdin);
	p = strchr(home_dir, '\n');
	if (p) *p = '\0';
      }
  }

  get_shell_name(shell_name);

  if (remove_flag)
    {
      remove_registry_data(personal_flag, version_specific_flag);
    }
  else
    {
      set_registry_data(personal_flag, version_specific_flag,
			main_root, shell_name, home_dir);
      if (!personal_flag)
	call_dumpprog(main_root);

      register_to_shell(main_root);
    }

  return 0;
}
